/*******************************************************
*   Normal Mode Analysis                               *
*******************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "ct2d.h"
/*** Basic  ***/
#include "Basic.h"
/*** Utility ***/
#include "LIButility.h"
/*** Truss Elements ***/
#include "LIBt2.h"
/*** VTF output **/
#include "writeVTF.h"

/*** static variables ***/
static int numnp;
static int nume;
static int numat;
static int nsect;
static int nspc;
static int ncload;
static NODE  *node;/**** Node Object ****/
static TRUSS *truss;/**** Truss element Object ****/
static MATERIAL *material;/**** material Object ****/
static SECTION  *section;/**** section Object ****/
static SPC   *spc;  /**** spc Object ****/
static CLOAD *cload;/**** cload Object ****/
static int nfreedom;/*** number of freedoms ***/
static int ncon;/*** number of constraint dofs ***/
static int *doftab;/*** doftable ***/
static double *tempx;/***  solution vector in condensed form ***/
static double *temp;/**** solution vector ****/
static double *cval;/** constrained value ***/
static RESULT_NODE    *result_node;/*** results node ***/
static RESULT_ELEMENT *result_element;/*** results element ***/
/*** static functions ***/
static void set_load();
static void assemble(FILE*);
static void add_gmat(int*,double*,int);
static void massmat(double* mvec);/** NormalModeAnalysis 2013dec **/
static void set_temp();
static void show_displacement(FILE*);
//static void show_rforce(FILE*);
//static void add_force(FILE*,int,double*);
//static void show_stress(FILE*);
static void writePLT(FILE*,FILE*,CT2D*);
/*** static variable ***/
static int kb;/** size of Number of subdiagonals **/
static double *band;/** band marix **/
static double *mvec;/** NormalModeAnalysis 2013dec **/
/** DPBSV Interface **/
#if LOWER
static char uplo='L';
#else
static char uplo='U';
#endif
static int  nrhs=1;
static int  ld_band;
static int  ld_rhs;
static int  info;
/*** LAPACK DPBSV Solver  **/
extern void dpbsv_();
/*** Prof. Samukawa's solver **/
/** for Lower Triangle Matrix **/
/*******************************
  Normal Mode Analysis
********************************/
void NormalMode( FILE *fp_out, FILE *fp_log, FILE *fp_vtf, FILE *fp_plt, FILE *fp_mat, CT2D *ct2d)  // 2014apr04  add FILE *fp_mat
{
	int i, j, itr[2]={0,0};
	double tol, scal;
	double *work;
	double *amat, *evec, *v;
	int *ind;
	fprintf( fp_out,"#########################\n" );
	fprintf( fp_out,"#### Normal Mode Analysis\n" );
	fprintf( fp_out,"#########################\n" );
/**** Allocate array ***/ 
	ct2d->temp=vectorR( NDOF * (ct2d->numnp) );
	ct2d->tempx=vectorR( ct2d->nfreedom);
	if ( ( ct2d->result_node=(RESULT_NODE * )malloc( (ct2d->numnp) * sizeof( RESULT_NODE ) ) ) == NULL ) {
		fprintf(fp_log,"Error:Memory is not enough! a in result_node \n");
		exit(1);
	}
	if ( ( ct2d->result_element=(RESULT_ELEMENT * )malloc( (ct2d->nume) * sizeof( RESULT_ELEMENT ) ) ) == NULL ) {
		fprintf(fp_log,"Error:Memory is not enough! a in result_element \n");
		exit(1);
	}
/**** Set static variables ***/
	numnp =ct2d->numnp;
	nume  =ct2d->nume;
	numat =ct2d->numat;
	nsect =ct2d->nsect;
	nspc  =ct2d->nspc;
	ncload=ct2d->ncload;
	node =ct2d->node;
	truss=ct2d->truss;
	material=ct2d->material;
	section=ct2d->section;
	spc=ct2d->spc;
	cload=ct2d->cload;
	nfreedom=ct2d->nfreedom;
	ncon=ct2d->ncon;
	doftab=ct2d->doftab;
	cval=ct2d->cval;
	result_node   =ct2d->result_node;
	result_element=ct2d->result_element;
/** size of band matrix **/
	printf(" before find_bw nfreedom=%d\n",nfreedom);
	kb=find_bw(fp_out,fp_log,ct2d);
/** allocate bandmatrix **/
	band=vectorR( (kb+1)* nfreedom );
	mvec=vectorR( nfreedom );         /*  NormalModeAnalysis 2013dec */
	v=vectorR( nfreedom );     /*  2013dec  for sorted eigenvalues */
	ind=vectorI( nfreedom );   /*  2013dec  index vector for eigenvalues */
/** work **/
	tempx=ct2d->tempx;
	temp =ct2d->temp;
/** create & assemble stiffness matrix **/
	assemble( fp_log );
	printf(" assemble finish nfreedom=%d\n",nfreedom);
	massmat( mvec );         /*  NormalModeAnalysis 2013dec **/
	printf(" mass vector nfreedom=%d\n",nfreedom);
//	mvec[0]=mvec[0]*2;
	for(i=0; i<nfreedom; i++){
		printf("%d %e\n",i,mvec[i]);
	}
	printf(" massmat finish nfreedom=%d\n",nfreedom);

	amat=vectorR( nfreedom * nfreedom );
	evec=vectorR( nfreedom * nfreedom );
	genamat( mvec,band,amat,nfreedom,kb+1,fp_mat); /* NormalModeAnalysis 2013dec */


//	tol=0.00001; itr[0]=nfreedom*nfreedom*20;
//	tol=0.0000000001; itr[0]=nfreedom*nfreedom*200;
	tol=0.000000000000001; itr[0]=nfreedom*nfreedom*2000;
	jacobev (amat,evec,tol,nfreedom,itr );  /*  NormalModeAnalysis 2013dec */
	printf("iteration=%d\n", itr[1]);
	for (i=0; i<nfreedom; i++){
		v[i] = amat[i*nfreedom+i]; 
	}

	bsort(v,nfreedom,ind);
	for (i=0; i<nfreedom; i++){  printf("%20.16lf %20.16lf\n", v[i],amat[ind[i]*nfreedom+ind[i]]);}

	for (i=0; i<nfreedom; i++){  printf(" %d", ind[i]);}
	printf("\n");

	for (i=0; i<nfreedom; i++){  
		printf("%d-th eigenvalue %20.16lf\n",i,v[i]);
		for (j=0; j<nfreedom; j++){  
			scal = 1.0 / sqrt(mvec[j]);
			if(j%6==5){
				printf(" %10.5f\n", scal * evec[ind[i]*nfreedom+j]);
			}else{
				printf(" %10.5f", scal * evec[ind[i]*nfreedom+j]);
			}
		}
		printf("\n");
	}
/******************************
 -0.0000000000000010  0
 -0.0000000000000009  1
 -0.0000000000000003  2
  1.5542153421356324  3
  1.5542153421356331  4
  1.7764002763288915  
  2.1749571343190870  
  2.9551845001450321  
  3.3802689390754406  8
  3.3802689390754423  9
  3.4557340167662254 
  4.4875551994581215 11
  4.4875551994581251 12
  4.7531857373243422 
  5.0448154998549661 
  6.1385433842995578 
  8.1613063473505960 16
  8.1613063473506102 17
 10.7084618591232523 18
 10.7084618591232683 19
 10.8698583655358707
 11.2468142626756631
 11.6787124827986695
 12.2643183387914156
 12.9322698121317554 24
 12.9322698121317661 25
*******************************/
/** Select X-th eigenvector move to tempx ***/
	for (j=0; j<nfreedom; j++){  
		scal = 1.0 / sqrt(mvec[j]);
//		tempx[j]=scal * evec[ind[0]*nfreedom+j];   // 1-th vector
//		tempx[j]=scal * evec[ind[1]*nfreedom+j];   // S13Free1.ps
//		tempx[j]=scal * evec[ind[2]*nfreedom+j];   // S13Free2.ps
//		tempx[j]=scal * evec[ind[3]*nfreedom+j];   // S13Free3.ps
		tempx[j]=scal * evec[ind[4]*nfreedom+j];   // S13Free4.ps
//		tempx[j]=scal * evec[ind[5]*nfreedom+j];   // S13Free5.ps
//		tempx[j]=scal * evec[ind[6]*nfreedom+j];   // S13Free6.ps
//		tempx[j]=scal * evec[ind[7]*nfreedom+j];   // S13Free7.ps
//		tempx[j]=scal * evec[ind[8]*nfreedom+j];   // S13Free8.ps
//		tempx[j]=scal * evec[ind[9]*nfreedom+j];   // S13Free9.ps
//		tempx[j]=scal * evec[ind[11]*nfreedom+j];  // S13Free11.ps
//		tempx[j]=scal * evec[ind[12]*nfreedom+j];  // S13Free12.ps
//		tempx[j]=scal * evec[ind[16]*nfreedom+j];  // S13Free16.ps
//		tempx[j]=scal * evec[ind[17]*nfreedom+j];  // S13Free17.ps
//		tempx[j]=scal * evec[ind[18]*nfreedom+j];  // S13Free18.ps
//		tempx[j]=scal * evec[ind[19]*nfreedom+j];  // S13Free19.ps
//		tempx[j]=scal * evec[ind[24]*nfreedom+j];  // S13Free24.ps
//		tempx[j]=scal * evec[ind[25]*nfreedom+j];  // S13Free25.ps
	}
/** transform from tempx to temp ***/
	set_temp();
/** show displacement **/
	show_displacement( fp_out );

	fprintf( fp_out,"##### Eigenvalues ###############\n" );
	fprintf( fp_out," tol=%e\n",tol);
	for (i=0; i<nfreedom; i++){  fprintf( fp_out,"%20.16lf\n", v[i]);}

//  	for (j=0; j<nfreedom; j++){
//		for (i=j*nfreedom; i<j*nfreedom+nfreedom; i++){  printf("%20.16lf\n", amat[i]);}
//	}

	free_vectorR (amat);
	free_vectorR (band);
	free_vectorR (mvec);
	free_vectorR (v);
//	free_vectorR (work);
/*** freeing allocated memory ***/ 
	free_vectorR(temp);
	free_vectorR(tempx);
/*** show rforce ***/
//	show_rforce( fp_out);
/*** show stress ***/
//	show_stress( fp_out );
/*** write vtf ***/
//	writeVTF( fp_log, fp_vtf, ct2d);
/*** write plot ***/
	writePLT( fp_log, fp_plt, ct2d);
/*** message **/
	fprintf( fp_out,"### NormalModeAnalysis FINISHED!\n");
	fprintf( fp_log,"### NormalModeAnalysis FINISHED!\n");
	fflush( fp_log );
}


/*************************
 assemble lumped mass matrix
**************************/
static void massmat(double* mvec)
{
	int nid[NNOD];
	int i,j,k,nn,ix;
	double xx[NNOD],yy[NNOD];
	double rho,area,length;
	int mid;
	int sid;
/*** truss *****/
	nn=2;
	for(i=0;i<nume;i++){
//		printf("massmat i=%d\n",i);
		for(j=0;j<nn;j++){
			nid[j]=truss[i].id[j];
//			printf("nid[%d]=%d\n",j,nid[j]);
		}
/** coordinate of 2-node **/
		for(j=0;j<nn;j++){
			xx[j]=node[nid[j]-1].cord.x;
			yy[j]=node[nid[j]-1].cord.y;
		}
/** section **/
		sid=truss[i].sid;
		area=section[sid-1].area;
		mid =section[sid-1].mid;
/** material **/
//		young=material[mid-1].young;	
		rho=1.0;	
		length=sqrt((xx[1]-xx[0])*(xx[1]-xx[0])+(yy[1]-yy[0])*(yy[1]-yy[0]));
		for(k=0;k<NDOF;k++){
			ix=doftab[NDOF*(nid[0]-1)+k];
//			printf("nid[0] ix=%d\n",ix);
			if(ix>0){ mvec[ix-1] += area*rho*length*0.5;}
		}
		for(k=0;k<NDOF;k++){
			ix=doftab[NDOF*(nid[1]-1)+k];
//			printf("nid[1] ix=%d\n",ix);
			if(ix>0){ mvec[ix-1] += area*rho*length*0.5;}
		}
	}
//	fprintf(fp,"### assemble finished!\n");
}

/*************************
 assemble global matrix
**************************/
static void assemble(FILE *fp)
{
	double stf[NNOD*NDOF*(NNOD*NDOF+1)/2];
	int nid[NNOD],iwk[NNOD*NDOF];
	int i,j,k,nn;
	int head;
	double xx[NNOD],yy[NNOD];
	double young,area;
	int mid;
	int sid;
/*** truss *****/
	nn=2;
	for(i=0;i<nume;i++){
		for(j=0;j<nn;j++){
			nid[j]=truss[i].id[j];
		}
/** coordinate of 2-node **/
		for(j=0;j<nn;j++){
			xx[j]=node[nid[j]-1].cord.x;
			yy[j]=node[nid[j]-1].cord.y;
		}
/** section **/
		sid=truss[i].sid;
		area=section[sid-1].area;
		mid =section[sid-1].mid;
/** material **/
		young=material[mid-1].young;	
/** head **/
		head=0;
		for(j=0;j<nn;j++){
			for(k=0;k<NDOF;k++){
				iwk[ head + k ]=doftab[NDOF*(nid[j]-1)+k];
			}
			head+=NDOF;
		}
/*** 2-node Element **/
		mk_t2(fp,xx,yy,young,area,stf);
/***add global stiffness *****/
		add_gmat(iwk,stf,head);
	}
	fprintf(fp,"### assemble finished!\n");
}
/****************************
   add global stiffness    
****************************/
static void add_gmat(int iwk[], double stf[], int nn)
{   
	int ii,jj,ix,jx,num;
	double sss;
/** loop over all dofs **/
	for(jj=0;jj<nn;jj++){
		jx=iwk[ jj ];
/* OFF DIAGONAL TERM */
		for( ii=0;ii<jj;ii++){
			num=jj*(jj+1)/2+ii;
			ix =iwk[ ii ];
			sss=stf[ num ];
			if( ix > 0 && jx > 0 ){
/* BOTH INDEPENDED */
				if( jx > ix ) {
#if LOWER
					band[ (kb+1)*(ix-1) + jx-ix ]+=sss;
#else
					band[ (kb+1)*(jx-1) + kb+ix-jx ]+=sss;
#endif
				}else{
#if LOWER
					band[ (kb+1)*(jx-1) + ix-jx ]+=sss;
#else
					band[ (kb+1)*(ix-1) + kb+jx-ix ]+=sss;
#endif
				}
			}else if( ix > 0 && jx < 0 ){
/* CONSIDERING CONSTRAINT CONDITION FOR JX */
				tempx[ix-1]+=-sss*cval[-jx-1];
			}else if( ix < 0 && jx > 0 ){
/* CONSIDERING CONSTRAINT CONDITION FOR IX */
				tempx[jx-1]+=-sss*cval[-ix-1];
			}else if( ix < 0 && jx < 0 ){
/* NOTHING TO DO */
			}
		}
/* DIAGONAL TERM */
		num=(jj+1)*(jj+2)/2-1;
		sss=stf[num];
		if( jx > 0 ){
#if LOWER
			band[ (kb+1)*(jx-1) ]+=sss;
#else
			band[ (kb+1)*(jx-1) + kb ]+=sss;
#endif
		}else if( jx < 0 ){
/* NOTHING TO DO */
		}
	}
}
/***************************
 Transform from tempx to temp
****************************/
static void set_temp( FILE* fp )
{
	int i,j,jj;
	double val;
/** Loopover nodes ***/
	for(i=0;i<numnp;i++){
		for(j=0;j<NDOF;j++){
			jj=doftab[NDOF*i+j];
			if( jj > 0 ){
				temp[NDOF*i+j]=tempx[jj-1];
			}else if( jj < 0 ){
				val=cval[-jj-1];
				temp[NDOF*i+j]=val;
			}
		}
	}
}
/*********************
	Show Displacement   
**********************/
static void show_displacement( FILE* fp)
{
	int i;
	fprintf(fp,"*** Displacement (U1/U2)****\n");
	for(i=0;i<numnp;i++){
		fprintf(fp,"%5d %10.4e %10.4e\n",i+1,temp[2*i],temp[2*i+1]);
		result_node[i].u[0]=temp[2*i  ];
		result_node[i].u[1]=temp[2*i+1];
	}
}
/***************************************************
    write PLOT file
***************************************************/
static void writePLT(FILE* fp_log, FILE* fp_plt, CT2D *ct2d)
{
	int i;
/** title ***/
	fprintf(fp_plt,"%s\n",ct2d->title);
/** Number of nodes, elements ***/
	fprintf(fp_plt,"%d %d\n",ct2d->numnp,ct2d->nume);
/** nodes ***/
	for(i=0;i<ct2d->numnp;i++){
		fprintf(fp_plt,"%e %e\n",(ct2d->node)[i].cord.x,(ct2d->node)[i].cord.y);
	}
/** elements ***/
	for(i=0;i<ct2d->nume;i++){
		fprintf(fp_plt,"%d %d\n",
		(ct2d->truss)[i].id[0],
		(ct2d->truss)[i].id[1]);
	}
/** Results (displacement) ***/
	for(i=0;i<ct2d->numnp;i++){
		fprintf(fp_plt,"%e %e\n",
		(ct2d->result_node)[i].u[0],
		(ct2d->result_node)[i].u[1]);
	}
/** Results (stress) ***/
	for(i=0;i<ct2d->nume;i++){
		fprintf(fp_plt,"%e\n",(ct2d->result_element)[i].s11);
	}
	fflush( fp_plt );
}

